import json
import logging
import uuid
from typing import Iterable, TypedDict, get_args

from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models
from django.utils.functional import cached_property
from django.utils.text import slugify
from ordered_model.models import OrderedModel

from . import formkit_schema

logger = logging.getLogger()


class UuidIdModel(models.Model):
    """
    Consistently use fields which will
    help with syncing data:
     - UUID field is the ID
     - Created field
     - Last Modified field
     - updated_by (optional)
     - created_by (optional)
    """

    class Meta:
        abstract = True

    id = models.UUIDField(primary_key=True, default=uuid.uuid4)
    created = models.DateTimeField(auto_now_add=True, blank=True, null=True)
    updated = models.DateTimeField(auto_now=True, blank=True, null=True)
    created_by = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.PROTECT, related_name="+", blank=True, null=True
    )
    updated_by = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.PROTECT, related_name="+", blank=True, null=True
    )


class OptionDict(TypedDict):
    value: str
    label: str


class Option(OrderedModel, UuidIdModel):
    """
    This is a key/value field representing one "option" for a FormKit property
    The translated values for this option are in the `Translatable` table
    """

    value = models.CharField(max_length=1024)
    label = models.CharField(max_length=1024)
    field = models.ForeignKey(
        "FormKitSchemaNode",
        on_delete=models.CASCADE,
        limit_choices_to={"node__formkit__in": ["select", "radio"]},
    )
    order_with_respect_to = "field"

    @classmethod
    def from_pydantic(cls, options: list[str] | list[OptionDict]) -> Iterable["Option"]:
        for option in options:
            if isinstance(option, str):
                yield cls(value=option, label=option)
            elif isinstance(option, dict) and option.keys() == {"value", "label"}:
                yield cls(**option)

    def __str__(self):
        return f"{self.label} : {self.value}"


class FormComponents(OrderedModel, UuidIdModel):
    """
    A model relating "nodes" of a schema to a schema with model ordering
    """

    schema = models.ForeignKey("FormKitSchema", on_delete=models.CASCADE)
    node = models.ForeignKey("FormKitSchemaNode", on_delete=models.CASCADE)
    label = models.CharField(max_length=1024, help_text="Used as a human-readable label")

    order_with_respect_to = "schema"

    class Meta:
        ordering = ("schema", "order")

    def __str__(self):
        return f"{self.node}[{self.order}]: {self.schema}"


class Membership(OrderedModel, UuidIdModel):
    """
    This is an ordered m2m model representing
    how parts of a "FormKit group" are arranged
    """

    group = models.ForeignKey(
        "FormKitSchemaNode",
        on_delete=models.CASCADE,
        limit_choices_to={"node__formkit": "group"},
    )
    member = models.ForeignKey("FormKitSchemaNode", on_delete=models.CASCADE, related_name="members")
    order_with_respect_to = "group"


class NodeChildren(OrderedModel, UuidIdModel):
    """
    This is an ordered m2m model representing
    the "children" of an HTML element
    """

    parent = models.ForeignKey(
        "FormKitSchemaNode",
        on_delete=models.CASCADE,
        limit_choices_to={"node__node_type": "$el"},
    )
    child = models.ForeignKey("FormKitSchemaNode", on_delete=models.CASCADE, related_name="parent")
    order_with_respect_to = "parent"

    class Meta:
        ordering = ("order",)


class FormKitSchemaNode(UuidIdModel):
    """
    This represents a single "Node" in a FormKit schema.
    There are several different types of node which may be defined:
    FormKitSchemaDOMNode
    | FormKitSchemaComponent
    | FormKitSchemaTextNode
    | FormKitSchemaCondition
    | FormKitSchemaFormKit
    """

    NODE_TYPE_CHOICES = (
        ("$cmp", "Component"),
        ("text", "Text"),
        ("condition", "Condition"),
        ("$formkit", "FormKit"),
        ("$el", "Element"),
        ("raw", "Raw JSON"),
    )
    FORMKIT_CHOICES = [(t, t) for t in get_args(formkit_schema.FORMKIT_TYPE)]

    ELEMENT_TYPE_CHOICES = [("p", "p"), ("h1", "h1"), ("h2", "h2"), ("span", "span")]
    node_type = models.CharField(max_length=256, choices=NODE_TYPE_CHOICES, blank=True, help_text="")
    description = models.CharField(
        max_length=4000,
        null=True,
        blank=True,
        help_text="Decribe the type of data / reason for this component",
    )
    label = models.CharField(max_length=1024, unique=True, help_text="Used as a human-readable label")
    group = models.ManyToManyField("self", through=Membership, blank=True)
    children = models.ManyToManyField("self", through=NodeChildren, blank=True)

    node = models.JSONField(
        null=True,
        blank=True,
        help_text="A JSON representation of select parts of the FormKit schema",
    )

    additional_props = models.JSONField(
        null=True,
        blank=True,
        help_text="User space for additional, less used props",
    )

    translation_context = models.CharField(
        max_length=1024,
        null=True,
        blank=True,
        help_text="The gettext context to use to translate Options, label, placeholder and help text",
    )

    def __str__(self):
        return self.label

    def save(self, *args, **kwargs):
        """
        On save validate the 'node' field matches the 'FormKitNode'
        """
        return super().save(*args, **kwargs)

    @cached_property
    def node_options(self):
        """
        Because "options" are translated and
        separately stored, this step is necessary to
        reinstate them
        """
        options: Iterable[Option] = self.option_set.all()
        return [{"value": option.value, "label": option.label} for option in options]

    def get_node_values(self) -> dict:
        """
        Reify a 'dict' instance suitable for creating
        a FormKit Schema node from
        """
        if not self.node:
            return {}
        values = {**self.node}
        if self.node_options:
            values["options"] = self.node_options
        return values

    def get_node(self) -> formkit_schema.Node:
        """
        Return a "decorated" node instance
        with restored options and translated fields
        """
        node_content = self.get_node_values()
        formkit_node = formkit_schema.FormKitNode.parse_obj(node_content)
        # 'FormKitNode' is really a container to store any kind
        # of Node
        return formkit_node.__root__

    @classmethod
    def from_pydantic(
        cls, input_models: Iterable[formkit_schema.FormKitNode]
    ) -> Iterable[tuple["FormKitSchemaNode", Iterable[Option]]]:
        for input_model in input_models:
            instance = cls()
            if options := getattr(input_model, "options", None):
                option_models = Option.from_pydantic(options)
            else:
                option_models = Option.objects.none()
            instance.node = input_model.dict(exclude={"options"})
            # All other properties are passed directly
            yield instance, option_models


class SchemaManager(models.Manager):
    """
    Provides prefetching which we'll almost always want to have
    """

    def get_queryset(self):
        return super().get_queryset().prefetch_related("nodes", "nodes__children")


class FormKitSchema(UuidIdModel):
    """
    This represents a "FormKitSchema" which is an heterogenous
    collection of items.
    """

    key = models.TextField(max_length=1024, unique=True)
    nodes = models.ManyToManyField(FormKitSchemaNode, through=FormComponents)
    objects = SchemaManager()

    def get_schema_values(self):
        """
        Return a list of "node" dicts
        """
        for node in self.nodes.all():
            yield node.get_node_values()

    def to_pydantic(self):
        values = list(self.get_schema_values())
        return formkit_schema.FormKitSchema.parse_obj(values)

    def __str__(self):
        return self.key

    def clean(self):
        slugified = slugify(self.key)
        for instance in self._meta.model.objects.exclude(pk=self.pk):
            if slugified == slugify(instance.key):
                raise ValidationError(
                    f"The name '{self.key} clashed with {instance.key}. Please enter a different name."
                )
        return super().clean()

    @classmethod
    def from_pydantic(cls, input_model: formkit_schema.FormKitSchema) -> "FormKitSchema":
        """
        Converts a given Pydantic representation of a Schema
        to Django database fields
        """
        instance = cls.objects.create()
        for node, options in FormKitSchemaNode.from_pydantic(input_model.__root__):
            node.save()
            if options:
                for option in options:
                    option.field = node
                    option.save()
            FormComponents.objects.create(schema=instance, node=node)
            logger.info("Schema load from JSON done")
        return instance

    @classmethod
    def from_json(cls, input_file: json):
        """
        Converts a given JSON string to a suitable
        Django representation
        """
        schema_instance = formkit_schema.FormKitSchema.parse_obj(input_file)
        return cls.from_pydantic(schema_instance)



from __future__ import annotations

from keyword import iskeyword
from typing import Iterable, Literal

from formkit_ninja.formkit_schema import FormKitSchemaFormKit, FormKitSchemaProps, GroupNode, RepeaterNode
from formkit_ninja.parser.logger import log


class NodePath:
    """
    Mostly a wrapper around "tuple" to provide useful conventions
    for naming
    """

    def __init__(self, *nodes: tuple[GroupNode | RepeaterNode | FormKitSchemaProps]):
        self.nodes = nodes

    def __truediv__(self, node: Literal[".."] | GroupNode | RepeaterNode | FormKitSchemaProps):
        """
        This overrides the builtin '/' operator, like "Path", to allow appending nodes
        """
        if node == "..":
            return NodePath(*self.nodes[:-1])
        return NodePath(*self.nodes, node)

    def suggest_model_name(self) -> str:
        """
        Single reference for table name and foreign key references
        """
        model_name = "".join(map(self.safe_node_name, self.nodes))
        return model_name

    def suggest_class_name(self):
        model_name = "".join(map(lambda n: n.capitalize(), map(self.safe_node_name, self.nodes)))
        return model_name

    def suggest_field_name(self):
        """
        Single reference for table name and foreign key references
        """
        return self.safe_node_name(self.node)

    def suggest_link_class_name(self):
        return f"{self.suggest_class_name()}Link"

    @staticmethod
    def safe_name(name: str) -> str:
        if name is None:
            raise TypeError
        if not name.isidentifier() or iskeyword(name):
            raise KeyError(f"The name:  '''{name}''' is not a valid identifier")
        return name

    def safe_node_name(self, node: FormKitSchemaFormKit) -> str:
        """
        Return either the "name" or "id" field
        """
        if node.name:
            name = self.safe_name(node.name)
        elif node.html_id:
            name = self.safe_name(node.html_id)
        else:
            raise AttributeError("Could not determine a suitable 'name' for this node")

        return name

    @property
    def is_repeater(self):
        return isinstance(self.node, RepeaterNode)

    @property
    def is_group(self):
        return isinstance(self.node, GroupNode)

    @property
    def formkits(self) -> Iterable["NodePath"]:
        for n in self.children:
            if hasattr(n, "formkit"):
                yield self / n

    @property
    def formkits_not_repeaters(self) -> Iterable["NodePath"]:
        for n in self.children:
            if hasattr(n, "formkit") and not isinstance(n, RepeaterNode):
                yield self / n

    @property
    def children(self):
        return getattr(self.node, "children", []) or []

    def filter_children(self, type_) -> Iterable["NodePath"]:
        for n in self.children:
            if isinstance(n, type_):
                yield self / n

    @property
    def repeaters(self):
        yield from self.filter_children(RepeaterNode)

    @property
    def groups(self):
        yield from self.filter_children(GroupNode)

    @property
    def node(self):
        return self.nodes[-1]

    @property
    def parent(self):
        if len(self.nodes) > 1:
            return self.nodes[-2]
        else:
            return None

    @property
    def is_child(self):
        return self.parent is not None

    @property
    def depth(self):
        return len(self.nodes)

    def tail(self):
        return NodePath(self.node)

    def __str__(self):
        return f"NodePath {len(self.nodes)}: {self.node}"


class ToPydantic:
    def __call__(self, nodes: NodePath) -> Literal["str", "int", "bool", "Decimal", "float", "date"] | str:
        node = nodes.node

        if node.formkit == "number":
            if node.step is not None:
                # We don't actually **know** this but it's a good assumption
                return "float"
            return "int"

        match node.formkit:
            case "text":
                return "str"
            case "number":
                log(
                    f"[red]Underspecified field {node.formkit} field {node.name}. Could be int, float, or decimal. Defaulting to float."
                )
                return "float"
            case "select" | "dropdown" | "radio" | "autocomplete":
                log(f"[red]Underspecified {node.formkit} field {node.name}. Defaulting to string.")
                return "str"
            case "datepicker":
                return "datetime"
            case "tel":
                return "int"
            case "group":
                return nodes.suggest_class_name()
            case "repeater":
                return f"list[{nodes.suggest_class_name()}]"

        log(f"[yellow]Unknown field {node.formkit} field {node.name}. Defaulting to string.")
        return "str"


postgres_type = Literal["int", "text", "boolean", "NUMERIC(15,2)"]
django_type = Literal["ForeignKey", "DateField", "DateTimeField", "DecimalField", "ForeignKey", "OneToOneField"]


def to_postgres(nodes: NodePath) -> postgres_type:
    """
    Returns a suitable Postgres field type for data coerced from JSON
    """

    match ToPydantic()(nodes):
        case "bool":
            return "boolean"
        case "str":
            return "text"
        case "Decimal":
            return "NUMERIC(15,2)"
        case "int":
            return "int"
        case "float":
            return "float"

    return "text"


class ToDjango:
    """
    Returns a string suitable for a Django models file `field` parameter
    """

    def __call__(self, nodes: NodePath, config: "ParserConfig" | None = None) -> tuple[django_type, tuple[str, ...]]:
        if nodes.is_group:
            log(f"[yellow]Group node: {nodes.safe_node_name(nodes.node)}")
            return "OneToOneField", (
                nodes.suggest_class_name(),
                "on_delete=models.CASCADE",
            )
        if not config:
            config = ParserConfig
        match config.PydanticParser()(nodes):
            case "bool":
                return "BooleanField", ("null=True", "blank=True")
            case "str":
                return "TextField", ("null=True", "blank=True")
            case "Decimal":
                return "DecimalField", ("max_digits=20", "decimal_places=2", "null=True", "blank=True")
            case "int":
                return "IntegerField", ("null=True", "blank=True")
            case "float":
                return "FloatField", ("null=True", "blank=True")
            case "datetime":
                return "DateTimeField", ("null=True", "blank=True")
            case "date":
                return "DateField", ("null=True", "blank=True")


class BaseDjangoAttrib:
    def __init__(self, fieldname: str, fieldtype: str, args: tuple[str]):
        self.fieldname = fieldname
        self.fieldtype = fieldtype
        self.args = args or tuple()

    def __repr__(self):
        return f"{self.fieldname}, {self.fieldtype}, {self.args}"

    def __iter__(self):
        try:
            fieldname = NodePath.safe_name(self.fieldname)
        except TypeError:
            raise
        argslist = ", ".join(self.args) if self.args else ""
        yield f"    {fieldname} = models.{self.fieldtype}({argslist})"


class DjangoAttrib(BaseDjangoAttrib):
    """
    A factory to generate a Django model attribute from a
    Formkit node
    """

    def __init__(self, nodes: NodePath | FormKitSchemaFormKit, config: "ParserConfig" | None = None):
        """
        A single property
        """
        if not config:
            config = ParserConfig
        fieldtype, field_args = config.DjangoParser()(nodes, config)
        # The field name is based on the "last" node in the aequence
        super().__init__(fieldname=nodes.tail().suggest_model_name(), fieldtype=fieldtype, args=field_args)


class ParserConfig:
    DjangoParser = ToDjango
    PydanticParser = ToPydantic
    DjangoAttribParser = DjangoAttrib


class DjangoClassFactory:
    """
    A factory to generate a Django class definition
    """

    def __init__(
        self,
        nodes: NodePath,
        extra_attribs: list[BaseDjangoAttrib | str] | None = None,
        config: ParserConfig | None = None,
    ):
        self.config = config if config else ParserConfig()
        self.nodes = nodes
        self.extra_attribs = extra_attribs

    def __iter__(self):
        for r in self.nodes.repeaters:
            yield from iter(self.__class__(r, config=self.config))
        for g in self.nodes.groups:
            yield from iter(self.__class__(g, config=self.config))

        yield f"class {self.nodes.suggest_class_name()}(models.Model):"

        has_attributes = False

        # if a Repeater node, we want to preserve insertion order
        if self.nodes.is_repeater:
            has_attributes = True
            parent_class_name = (self.nodes / "..").suggest_class_name()
            related_name = (
                self.nodes.node.name
            )  # This is the property name in JSON. It's important to have for importers to know where to put nested data.
            yield "    # This class is a Repeater: Parent and ordinality fields have been added"
            yield f'    parent = models.ForeignKey("{parent_class_name}", on_delete=models.CASCADE, related_name="{related_name}")'
            yield "    ordinality = models.IntegerField()"

        if self.extra_attribs:
            has_attributes = True
            for e_a in self.extra_attribs:
                if isinstance(e_a, str):
                    yield e_a
                else:
                    yield from iter(e_a)
        for a in self.nodes.formkits:
            has_attributes = True
            # Skip "repeaters" as they are a separate model
            if a.is_repeater:
                continue
            yield from iter(DjangoAttrib(a, config=self.config))
        if not has_attributes:
            yield "    pass"

    @staticmethod
    def header():
        yield "from django.db import models\n"


class PydanticAttrib:
    """
    A factory to generate a Pydantic model attribute from a
    Formkit node
    """

    def __init__(self, nodes: NodePath | FormKitSchemaFormKit, opt: bool = True, config: ParserConfig | None = None):
        self.nodes = nodes
        self.opt = opt
        self.config = config or ParserConfig

    def __iter__(self):
        fieldtype = self.config.PydanticParser()(self.nodes)
        fieldname = self.nodes.suggest_field_name()
        yield f"    {fieldname}: {fieldtype} {'| None = None' if self.opt else ''}"


class PydanticClassFactory:
    def __init__(
        self,
        nodes: NodePath,
        extra_attribs: list[PydanticAttrib | str] | None = None,
        config: ParserConfig | None = None,
        klassname="BaseModel",  # Allows us to use a Django-Ninja schema or a subclass of models.Model
    ):
        self.nodes = nodes
        self.extra_attribs = extra_attribs
        self.config = config or ParserConfig
        self.klassname = klassname

    def get_extra_attribs(self):
        return self.extra_attribs

    def __iter__(self) -> Iterable[str]:
        """
        Returns a string generator to generate a complete
        set of Pydantic classes
        """
        # Recursively write dependencies
        for n in self.nodes.repeaters:
            yield from self.__class__(n, config=self.config)
        for g in self.nodes.groups:
            yield from self.__class__(g, config=self.config)

        # Write the class for the "current" node

        yield f"\nclass {self.nodes.suggest_class_name()}({self.klassname}):"
        has_attributes = False
        # Attributes
        if self.extra_attribs:
            has_attributes = True
            for e_a in self.extra_attribs:
                if isinstance(e_a, str):
                    yield e_a
                else:
                    yield from iter(e_a)

        for c in self.nodes.formkits:
            has_attributes = True
            yield from iter(PydanticAttrib(c, config=self.config))

        if not has_attributes:
            yield "    pass"
        else:
            yield "\n"

    @staticmethod
    def pydantic_header() -> Iterable[str]:
        yield "from datetime import datetime\n"
        yield "from decimal import Decimal\n"
        yield "from pydantic import BaseModel\n"
        yield "from typing import Union, Literal"


class PydanticSchemaClassFactory(PydanticClassFactory):
    """
    Subclasses `PydanticClassFactory` to prefer Django-Ninja's `schema`
    """

    def __init__(*args, **wkargs):
        super().__init__(klassname="Schema")

    @staticmethod
    def pydantic_header() -> Iterable[str]:
        yield from super().pydantic_header()
        yield "from ninja import Schema"


class DjangoAdminClassFactory(DjangoClassFactory):
    def _list_display(self):
        if attribs := list(self.nodes.formkits_not_repeaters):
            yield "    list_display = ["
            for a in attribs:
                yield f'        "{a.node.name}",'
            yield "    ]"

    # All fields are readonly
    def _readonly_fields(self):
        if attribs := list(self.nodes.formkits_not_repeaters):
            yield "    readonly_fields = ["
            for a in attribs:
                yield f'        "{a.node.name}",'
            yield "    ]"

    def _inlines(self):
        if reps := list(self.nodes.repeaters):
            yield "    inlines = ["
            for a in reps:
                yield f"        {a.suggest_class_name()}Inline,"
            yield "    ]"

    def __iter__(self):
        for r in self.nodes.repeaters:
            yield from iter(self.__class__(r, config=self.config))
        for g in self.nodes.groups:
            yield from iter(self.__class__(g, config=self.config))

        if self.nodes.is_repeater:
            yield f"class {self.nodes.suggest_class_name()}Inline(ReadOnlyInline):"
            yield f"    model = models.{self.nodes.suggest_class_name()}"

        yield f"@admin.register(models.{self.nodes.suggest_class_name()})"
        yield f"class {self.nodes.suggest_class_name()}Admin(admin.ModelAdmin):"
        yield from self._list_display()
        yield from self._inlines()
        yield from self._readonly_fields()

        # has_attributes = []

        # # if a Repeater node, we want to preserve insertion order

        # if self.nodes.is_repeater:
        #     has_attributes = True
        #     parent_class_name = (self.nodes / "..").suggest_class_name()
        #     related_name = self.nodes.node.name  # This is the property name in JSON. It's important to have for importers to know where to put nested data.
        #     yield f"    # This class is a Repeater: Parent and ordinality fields have been added"
        #     yield f'    parent = models.ForeignKey("{parent_class_name}", on_delete=models.CASCADE, related_name="{related_name}")'
        #     yield "    ordinality = models.IntegerField()"

        # if self.extra_attribs:
        #     has_attributes = True
        #     for e_a in self.extra_attribs:
        #         if isinstance(e_a, str):
        #             yield e_a
        #         else:
        #             yield from iter(e_a)
        # for a in self.nodes.formkits:
        #     has_attributes = True
        #     # Skip "repeaters" as they are a separate model
        #     if a.is_repeater:
        #         continue
        #     yield from iter(DjangoAttrib(a, config=self.config))
        # if not has_attributes:
        #     yield "    pass"

    @staticmethod
    def header():
        yield "from django.contrib import admin\n"
        yield "from . import models\n"

        yield "class ReadOnlyInline(admin.TabularInline):\n"
        yield "    def has_change_permission(self, request, obj=None):\n"
        yield "        return False\n"
        yield "    def has_add_permission(self, request, obj=None):\n"
        yield "        return False\n"
        yield "    def has_delete_permission(self, request, obj=None):\n"
        yield "        return False\n"



from __future__ import annotations

import logging
import operator
import warnings
from functools import reduce
from typing import Any, Literal, Optional

from django import forms
from django.contrib import admin
from django.http import HttpRequest
from ordered_model.admin import OrderedInlineModelAdminMixin, OrderedModelAdmin, OrderedTabularInline

from formkit_ninja import models

logger = logging.getLogger(__name__)


class ItemAdmin(OrderedModelAdmin):
    list_display = ("name", "move_up_down_links")


class JsonDecoratedFormBase(forms.ModelForm):
    """
    Adds additional fields to the admin where a model has a JSON field
    and some appropriate (tbc?) field parameters
    """

    # extra = forms.CharField(label="Extra", max_length=128, required=False)
    # hello_world = forms.CharField(widget=forms.NumberInput, required=False)

    # key is the name of a `models.JSONField` on the model
    # value is a list of fields to get/set in that JSON field
    _json_fields: dict[str, tuple[str]] = {"my_json_field": ("formkit", "description", "name", "key", "html_id")}

    def get_json_fields(self) -> dict[str, tuple[str]]:
        """
        Custom which json fields will be included in the return
        """
        return self._json_fields

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        instance = kwargs["instance"]
        for field, keys in self.get_json_fields().items():
            # Extract the dict of JSON values from the model instance if supplied
            values = {}
            if instance:
                values = getattr(instance, field, {}) or {}  # Don't allow none
            for key in keys:
                # The value, extracted from the JSON value in the database
                field_value = values.get(key, None)
                # The initial value of the admin form is set to the value of the JSON attrib
                if key in self.fields:
                    self.fields.get(key).initial = field_value
                else:
                    warnings.warn(f"Unassociated JSON field found: {key}")
        return

    def save(self, commit=True):
        """
        Updates the JSON field(s) from the fields specified in the `_json_fields` dict
        """

        for field, keys in self.get_json_fields().items():
            # Populate a JSON field in a model named "form"
            # from a set of standard form elements
            data = {}
            for key in keys:
                if cleaned_data := self.cleaned_data[key]:
                    data[key] = cleaned_data
        setattr(self.instance, field, data)
        return super().save(commit=commit)


class NewFormKitForm(forms.ModelForm):
    class Meta:
        model = models.FormKitSchemaNode
        fields = ("label", "translation_context", "node_type", "description")


class OptionForm(forms.ModelForm):
    class Meta:
        model = models.Option
        exclude = ()


class FormComponentsForm(forms.ModelForm):
    class Meta:
        model = models.FormComponents
        exclude = ()


class FormKitSchemaNodeOptionsInline(OrderedTabularInline):
    model = models.Option
    form = OptionForm
    readonly_fields = (
        # "id",
        "order",
        "move_up_down_links",
    )
    ordering = ("order",)
    extra = 0


class FormKitSchemaComponentInline(OrderedTabularInline):
    model = models.FormComponents
    readonly_fields = (
        "node",
        "created_by",
        "updated_by",
        "id",
        "order",
        "move_up_down_links",
    )
    ordering = ("order",)
    extra = 0


class FormKitNodeGroupForm(JsonDecoratedFormBase):
    class Meta:
        model = models.FormKitSchemaNode
        fields = (
            "label",
            "translation_context",
            "description",
        )

    _json_fields = {
        "node": (
            "formkit",
            "if_condition",
        )
    }

    formkit = forms.ChoiceField(required=False, choices=models.FormKitSchemaNode.FORMKIT_CHOICES, disabled=True)
    if_condition = forms.CharField(
        widget=forms.TextInput,
        required=False,
    )


class FormKitNodeForm(JsonDecoratedFormBase):
    """
    This is the most common component type: the 'FormKit' schema node
    """

    class Meta:
        model = models.FormKitSchemaNode
        fields = ("label", "translation_context", "description", "additional_props")

    # The `_json_fields["node"]` item affects the admin form,
    # adding the fields included in the `FormKitSchemaProps.__fields__.items` dict
    _json_fields = {
        "node": (
            "formkit",
            "description",
            "name",
            "key",
            "html_id",
            "if_condition",
            "options",
            "label",
            "placeholder",
            "help",
            # Validation of fields
            "validation",
            "validationLabel",
            "validationVisibility",
            "validationMessages",
            "prefixIcon",
            # Numeric forms
            "min",
            "max",
            "step",
        )
    }

    formkit = forms.ChoiceField(required=False, choices=models.FormKitSchemaNode.FORMKIT_CHOICES)
    name = forms.CharField(required=False)
    if_condition = forms.CharField(widget=forms.TextInput, required=False)
    key = forms.CharField(required=False)
    label = forms.CharField(required=False)
    placeholder = forms.CharField(required=False)
    help = forms.CharField(required=False)
    html_id = forms.CharField(
        required=False, help_text="Use this ID if adding conditions to other fields (hint: $get(my_field).value === 8)"
    )
    options = forms.CharField(
        required=False, help_text="Use this if adding Options using a JS function (hint: $get(my_field).value )"
    )
    validation = forms.CharField(required=False)
    validationLabel = forms.CharField(required=False)
    validationVisibility = forms.CharField(required=False)
    validationMessages = forms.JSONField(required=False)
    validationRules = forms.CharField(
        required=False, help_text="A function for validation passed into the schema: a key on `formSchemaData`"
    )
    prefixIcon = forms.CharField(required=False)

    # NumberNode props

    max = forms.IntegerField(required=False)
    min = forms.IntegerField(required=False)
    step = forms.IntegerField(required=False)

    def get_fields(self, request, obj: models.FormKitSchemaNode):
        """
        Customise the returned fields based on the type
        of formkit node
        """
        return super().get_fields(request, obj)


class FormKitNodeRepeaterForm(FormKitNodeForm):
    def get_json_fields(self) -> dict[str, tuple[str]]:
        return {
            "node": (
                *(super()._json_fields["node"]),
                "addLabel",
                "upControl",
                "downControl",
                "itemsClass",
                "itemClass",
            )
        }

    addLabel = forms.CharField(required=False)
    upControl = forms.BooleanField(required=False)
    downControl = forms.BooleanField(required=False)
    itemsClass = forms.CharField(required=False)
    itemClass = forms.CharField(required=False)
    max = forms.IntegerField(required=False)
    min = forms.IntegerField(required=False)


class FormKitTextNode(JsonDecoratedFormBase):
    class Meta:
        model = models.FormKitSchemaNode
        fields = (
            "label",
            "translation_context",
            "description",
        )

    _json_fields = {"node": ("content",)}
    content = forms.CharField(
        widget=forms.TextInput,
        required=True,
    )


class FormKitElementForm(JsonDecoratedFormBase):
    class Meta:
        model = models.FormKitSchemaNode
        fields = (
            "label",
            "translation_context",
            "description",
        )

    _skip_translations = {"label", "placeholder"}
    _json_fields = {"node": ("el", "name", "if_condition", "classes")}

    el = forms.ChoiceField(required=False, choices=models.FormKitSchemaNode.ELEMENT_TYPE_CHOICES)
    name = forms.CharField(
        required=False,
    )
    classes = forms.CharField(
        required=False,
    )
    if_condition = forms.CharField(
        widget=forms.TextInput,
        required=False,
    )


class FormKitConditionForm(JsonDecoratedFormBase):
    class Meta:
        model = models.FormKitSchemaNode
        # fields = '__all__'
        fields = (
            "label",
            "translation_context",
            "description",
        )

    _json_fields = {"node": ("if_condition", "then_condition", "else_condition")}

    if_condition = forms.CharField(
        widget=forms.TextInput,
        required=False,
    )
    then_condition = forms.CharField(
        max_length=256,
        required=False,
    )
    else_condition = forms.CharField(
        max_length=256,
        required=False,
    )


class FormKitComponentForm(JsonDecoratedFormBase):
    class Meta:
        model = models.FormKitSchemaNode
        fields = (
            "label",
            "translation_context",
            "description",
        )

    _json_fields = {"node": ("if_condition", "then_condition", "else_condition")}


class MembershipInline(OrderedTabularInline):
    model = models.Membership
    fk_name = "group"
    extra = 0
    readonly_fields = (
        "order",
        "move_up_down_links",
    )
    ordering = ("order",)


class MembershipComponentInline(admin.StackedInline):
    model = models.Membership
    fk_name = "member"
    extra = 0


class NodeChildrenInline(OrderedTabularInline):
    """
    Nested HTML elements
    """

    model = models.NodeChildren
    readonly_fields = (
        "order",
        "move_up_down_links",
    )
    ordering = ("order",)
    fk_name = "parent"
    extra = 0


class FormKitSchemaForm(forms.ModelForm):
    class Meta:
        model = models.FormKitSchema
        exclude = ("name",)


@admin.register(models.FormKitSchemaNode)
class FormKitSchemaNodeAdmin(OrderedInlineModelAdminMixin, admin.ModelAdmin):
    list_display = ("label", "id", "node_type")

    def get_inlines(self, request, obj: models.FormKitSchemaNode | None):
        if not obj:
            return []

        formkit_node_type = (obj.node or {}).get("formkit", None)

        if formkit_node_type == "group":
            return [
                MembershipInline,
            ]
        elif formkit_node_type in {"radio", "select"}:
            return [
                MembershipComponentInline,
                FormKitSchemaNodeOptionsInline,
            ]
        elif formkit_node_type:
            return [
                MembershipComponentInline,
            ]

        if obj.node_type == "$el":
            return [
                NodeChildrenInline,
            ]

        else:
            return []

    # Note that although overridden these are necessary
    inlines = [FormKitSchemaNodeOptionsInline, NodeChildrenInline, MembershipInline]

    def get_fieldsets(
        self, request: HttpRequest, obj: Optional[models.FormKitSchemaNode] = ...
    ) -> list[tuple[Optional[str], dict[str, Any]]]:
        if not getattr(obj, "node_type", None):
            warnings.warn("Expected a 'Node' with a 'NodeType' in the admin form")

        fieldsets: list[tuple[str, dict]] = []
        if not obj:
            return super().get_fieldsets(request, obj)
        if obj.node_type == "$formkit":
            fieldsets.append(
                (
                    "Field Validation",
                    {
                        "fields": (
                            "validation",
                            "validationLabel",
                            "validationVisibility",
                            "validationMessages",
                            "validationRules",
                        )
                    },
                )
            )
            # The "Repeater" has some specific fields
            # See these specified in FormKitNodeRepeaterForm
            if obj.node and obj.node.get("formkit", None) == "repeater":
                fieldsets.append(
                    (
                        "Repeater field properties",
                        {"fields": ("addLabel", "upControl", "downControl", "itemsClass", "itemClass")},
                    )
                )

        grouped_fields = reduce(operator.or_, (set(opts["fields"]) for _, opts in fieldsets), set())
        fieldsets.insert(
            0, (None, {"fields": [field for field in self.get_fields(request, obj) if field not in grouped_fields]})
        )
        return fieldsets

    def get_form(
        self,
        request: Any,
        obj: models.FormKitSchemaNode | None,
        change: None = None,
        **kwargs,
    ):
        # Handle different formkit node types
        formkit_forms = {"group": FormKitNodeGroupForm, "repeater": FormKitNodeRepeaterForm}

        # Handle different node types
        forms = {
            "condition": FormKitConditionForm,
            "$formkit": FormKitNodeForm,
            "$el": FormKitElementForm,
            "text": FormKitTextNode,
            "$cmp": FormKitComponentForm,
        }
        if not obj:
            return NewFormKitForm
        if node_type := getattr(obj, "node_type", None) is None:
            return super().get_form(request, obj, change, **kwargs)

        if node_type == "$formkit":
            formkit_node_type: Literal["group", "repeater", None] = (obj.node or {}).get("formkit", None)
            # Some special 'formkitnode' types have their own admin page
            if formkit_node_type in formkit_forms:
                return formkit_forms.get(formkit_node_type)

        if node_type in forms:
            return forms[node_type]

        warnings.warn(f"Expected: one of [{','.join(forms.keys())}]. Got {node_type}")
        return super().get_form(request, obj, change, **kwargs)


@admin.register(models.Option)
class OptionAdmin(OrderedModelAdmin):
    form = OptionForm
    readonly_fields = ("id",)
    list_display = ("value", "label", "field", "move_up_down_links")


@admin.register(models.FormKitSchema)
class FormKitSchemaAdmin(OrderedInlineModelAdminMixin, admin.ModelAdmin):
    form = FormKitSchemaForm

    def get_inlines(self, request, obj: models.FormKitSchema | None):
        """
        For a "new object" do not show the Form Components
        """
        if not obj:
            return []
        return [
            FormKitSchemaComponentInline,
        ]

    inlines = [
        FormKitSchemaComponentInline,
    ]


@admin.register(models.FormComponents)
class FormComponentsAdmin(OrderedModelAdmin):
    list_display = ("label", "schema", "node", "move_up_down_links")



from typing import Any, Callable, Iterable, NamedTuple

from formkit_schema import FormKitSchemaDOMNode, FormKitSchemaFormKit, GroupNode, RepeaterNode
from rich.console import Console

from . import type_convert
from .get_schemas import groups, process_node, repeaters

strings = tuple[str]

console = Console()
log = console.log


class TableRow(NamedTuple):
    node_name: str
    node_formkit: str
    field_type: str
    p_field_type: type_convert.postgres_type


table_row = tuple[str, str, str, type_convert.postgres_type]


def get_table(
    form_name: str, form_path: strings, nodes: Iterable[FormKitSchemaFormKit | FormKitSchemaDOMNode]
) -> Iterable[tuple[strings, list[TableRow]]]:
    rows: list[TableRow] = []

    for node in (n for n in nodes if not isinstance(n, (RepeaterNode, GroupNode, FormKitSchemaDOMNode))):
        field_type = type_convert.to_django(form_name, form_path, node, log)
        p_field_type = type_convert.to_postgres(form_name, form_path, node, log)
        rows.append(TableRow(node.name, node.formkit, str(field_type), p_field_type))

    # Repeaters and Groups
    for repeater in repeaters(nodes):
        yield from get_table(form_name, (*form_path, repeater.name), repeater.children)

    for group in groups(nodes):
        yield from get_table(form_name, (*form_path, group.name), group.children)

    # Yield the rows here after repeater + group
    yield (form_path, rows)


def tables(get_forms: Callable[[], Iterable[str, Any]]):
    for form_name, part in get_forms():
        nodes: list[FormKitSchemaFormKit] = [process_node(n) for n in part["children"]]
        yield from get_table(form_name, (part["title"],), nodes)



import warnings
from typing import List
from uuid import UUID

from django.shortcuts import get_object_or_404
from ninja import ModelSchema, Router

from formkit_ninja import formkit_schema, models

router = Router(tags=["FormKit"])


class FormKitSchemaIn(ModelSchema):
    class Config:
        model = models.FormKitSchema
        model_fields = "__all__"


class FormKitSchemaListOut(ModelSchema):
    class Config:
        model = models.FormKitSchema
        model_fields = ("id", "key")


class FormComponentsOut(ModelSchema):
    node_id: UUID
    schema_id: UUID

    class Config:
        model = models.FormComponents
        model_fields = ("label",)


class NodeChildrenOut(ModelSchema):
    class Config:
        model = models.NodeChildren
        model_fields = ("parent", "child", "order")


@router.get("list-schemas", response=list[FormKitSchemaListOut])
def get_list_schemas(request):
    return models.FormKitSchema.objects.all()


@router.get("list-nodes", response=dict[str, formkit_schema.Node], exclude_defaults=True, exclude_none=True)
def get_formkit_nodes(request):
    """
    Get all of the FormKit nodes in the database
    """
    response: dict[str, formkit_schema.Node] = {}
    for node in models.FormKitSchemaNode.objects.all():
        try:
            response[f"{node.pk}"] = node.get_node()
        except:  # noqa: E722
            warnings.warn(f"An unparseable FormKit node was hit at {node.pk}")
    return response


@router.get("list-related-nodes", response=list[NodeChildrenOut], exclude_defaults=True, exclude_none=True)
def get_related_nodes(request):
    """
    Get all of the FormKit nodes in the database
    """
    return models.NodeChildren.objects.all()


@router.get(
    "list-components", response=List[FormComponentsOut], exclude_defaults=True, exclude_none=True, by_alias=True
)
def get_components(request):
    values = models.FormComponents.objects.all()
    return values


@router.get(
    "schema/{schema_id}",
    response=formkit_schema.FormKitSchema,
    exclude_none=True,
    by_alias=True,
)
def get_schemas(request, schema_id: UUID):
    schema: models.FormKitSchema = get_object_or_404(models.FormKitSchema.objects, id=schema_id)
    model = schema.to_pydantic()
    return model


@router.get(
    "node/{node_id}",
    response=formkit_schema.FormKitNode,
    exclude_none=True,
    by_alias=True,
)
def get_node(request, node_id: UUID):
    node: models.FormKitSchemaNode = get_object_or_404(models.FormKitSchemaNode.objects, id=node_id)
    instance = node.get_node()
    return instance
